import processing.core.*; 
import processing.xml.*; 

import processing.serial.*; 

import java.applet.*; 
import java.awt.Dimension; 
import java.awt.Frame; 
import java.awt.event.MouseEvent; 
import java.awt.event.KeyEvent; 
import java.awt.event.FocusEvent; 
import java.awt.Image; 
import java.io.*; 
import java.net.*; 
import java.text.*; 
import java.util.*; 
import java.util.zip.*; 
import java.util.regex.*; 

public class NU32v2_serial_bootloader extends PApplet {

//////////////////////////////////////////////////////////////////////////////////////
////////////  Nick Marchuk  //////////////////////////////////////////////////////////
////////////  NU32v2 Bootloader Application //////////////////////////////////////////
////////////  January 2011 ///////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

// serial library
 
Serial myPort; 

// buttons and labels
TextButtons[] button;
TextButtons[] serialButtons;
PFont bigfont, smallfont;
String loadPath;
String loadPathLabel, serialPortLabel, connectedLabel;
String messageText;
int number_buttons;
int number_serial_buttons;

// boolean controls for program control
boolean HexOK;
boolean SerialOK;
boolean ConnectedOK;
boolean WaitForProgrammingResponse, SendBinary, WaitForProgrammingConfirm;

//////////////////////////////////////////////////////////////////////////////////////
////////////  setup  /////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

public void setup()
{
  size(585, 305);
  background(230, 230, 230);

  bigfont = loadFont("Gisha-Bold-24.vlw"); 
  smallfont = loadFont("Gisha-Bold-16.vlw");

  HexOK = false;
  SerialOK = false;
  ConnectedOK = false;
  WaitForProgrammingResponse = false;
  WaitForProgrammingConfirm = false;
  SendBinary = false;

  loadPathLabel = "No .hex selected, click 'Select .hex File'";
  serialPortLabel = "No serial port selected, select from 'Serial Ports'";
  connectedLabel = "Not connected, hold Reset, hold G6, release Reset, release G6";
  messageText = "(Disconnected)";

  // serial
  //println(Serial.list());
  number_serial_buttons = Serial.list().length;
  serialButtons = new TextButtons[number_serial_buttons];
  int index_serial = 0;
  for (index_serial = 0; index_serial < number_serial_buttons; index_serial++) {
    serialButtons[index_serial] = new TextButtons(220,135+index_serial*40,165,35,bigfont,Serial.list()[index_serial]); // x,y,w,h,font,text
  }

  // other buttons
  number_buttons = 3;
  button = new TextButtons[number_buttons];
  button[0] = new TextButtons(10,260,55,35,bigfont,"Exit"); // x,y,w,h,font,text
  button[1] = new TextButtons(10,135,185,35,bigfont,"Select .hex File"); // x,y,w,h,font,text
  //button[2] = new TextButtons(185,95,160,32,smallfont,"small font example"); // x,y,w,h,font,text
  //button[2] = new TextButtons(410,215,75,35,bigfont,"Erase"); // x,y,w,h,font,text
  //button[3] = new TextButtons(410,255,155,35,bigfont,"Blank Check"); // x,y,w,h,font,text
  button[2] = new TextButtons(410,135,115,35,bigfont,"Program"); // x,y,w,h,font,text
  //button[5] = new TextButtons(410,255,80,35,bigfont,"Verify"); // x,y,w,h,font,text
  //button[5] = new TextButtons(410,295,60,35,bigfont,"Run"); // x,y,w,h,font,text
}

//////////////////////////////////////////////////////////////////////////////////////
////////////  draw  //////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

public void draw()
{
  int index = 0;
  background(230, 230, 230);

  // status background and text
  fill(255);
  rect(5,5,575,87);
  fill(0);
  textFont(smallfont);
  text("Status",10,20);
  if (HexOK) {
    fill(15,190,30);
    text(loadPathLabel,10,42); // load path
  }
  else {
    fill(165,5,5);
    text(loadPathLabel,10,42); // load path
  }

  if (SerialOK) {
    fill(15,190,30);
    text(serialPortLabel,10,64); // serial port
  }
  else {
    fill(165,5,5);
    text(serialPortLabel,10,64); // serial port
  }

  if (ConnectedOK) {
    fill(15,190,30);
    text(connectedLabel,10,86); // connected
  }
  else {
    fill(165,5,5);
    text(connectedLabel,10,86); // connected
  }

  // serial buttons
  if (!SerialOK) {
    fill(255);
    rect(215,100,175,35+number_serial_buttons*40);
    fill(0);
    textFont(smallfont);
    text("Serial Ports",220,120);
    for (index = 0; index < number_serial_buttons; index++) {
      serialButtons[index].update();
      serialButtons[index].display();
    }
  }

  // buttons
  fill(255);
  rect(5,100,195,75);
  fill(0);
  textFont(smallfont);
  text("Compiled Code",10,120);
  for (index = 0; index < 2; index++) {
    button[index].update();
    button[index].display();
  }
  // bootloader control buttons
  if (HexOK&&SerialOK&&ConnectedOK) {
    fill(255);
    rect(405,100,175,75);
    fill(0);
    textFont(smallfont);
    text("Bootloader Functions",410,120);
    for (index = 2; index < number_buttons; index++) {
      button[index].update();
      button[index].display();
    }
  }

  // message background and text, progress bar
  fill(255);
  rect(5,180,195,75);
  fill(0);
  textFont(smallfont);
  text("Message",10,200);
  textFont(bigfont);
  fill(0);
  text(messageText,10,235);
}

//////////////////////////////////////////////////////////////////////////////////////
////////////  mouseClicked  //////////////////////////////////////////////////////////
////////////  If a button is selected,  //////////////////////////////////////////////
////////////  perform a function releated to the buttons name  ///////////////////////
//////////////////////////////////////////////////////////////////////////////////////

public void mousePressed() {
  try {
    String buttonText = null;
    int index = 0;
    boolean otherButton = false;
    boolean serialButton = false;

    // check all of the buttons
    for (index = 0; index < number_buttons; index++) {
      if (button[index].over()) {
        otherButton = true;
        buttonText = button[index].thetext;
      }
    }
    for (index = 0; index < number_serial_buttons; index++) {
      if (serialButtons[index].over()) {
        serialButton = true;
        otherButton = false;
        buttonText = serialButtons[index].thetext;
      }
    }

    //println(buttonText);

    if (buttonText == null) {
      // nothing, no button was clicked
    }
    else {
      if (otherButton) {
        /////////////////////////////// Select .hex File //////////////////////////////////////////////////////
        if (buttonText.equals("Select .hex File")) {
          loadPath = selectInput("Select your .hex file");  // Opens file chooser
          if (loadPath == null) {
            loadPathLabel = "No .hex selected";
            HexOK = false;
          }
          else {
            String filetype = loadPath.substring(loadPath.length()-4,loadPath.length());
            //println("file type = " + filetype);
            if (filetype.equals(".hex")) {
              if (loadPath.length() > 60) {
                loadPathLabel = "Using '..." + loadPath.substring(loadPath.length()-55,loadPath.length()) + "'";
              }
              else {
                loadPathLabel = "Using '" + loadPath;
              }
              HexOK = true;
            }
            else {
              loadPathLabel = "No .hex selected";
              HexOK = false;
            }
          }
        }
        /////////////////////////////// Erase //////////////////////////////////////////////////////
        if (buttonText.equals("Erase")) {
          // send erase character
          myPort.write('E');
        }
        /////////////////////////////// Blank Check //////////////////////////////////////////////////////
        if (buttonText.equals("Blank Check")) {
          // send blank check character
          myPort.write('B');
        }
        /////////////////////////////// Program //////////////////////////////////////////////////////
        if (buttonText.equals("Program")) {

          try {

            cursor(WAIT);

            // send erase character
            myPort.write('E');

            String lines[] = loadStrings(loadPath);
            //println("there are " + lines.length + " lines");

            /////////// make the binary file ///////////////
            byte[] gDataBfr = new byte[2250000];
            byte[] dataToSend = new byte[32];
            int gPeakAddr = 0;
            int i = 0;
            String LineData;
            int LineDataCount; // byte
            int LineAddr;
            int LineCode; // byte
            int DataCount; // byte
            int OutAddr = 0;
            int LongAddr;
            int Checksum;
            int Outoffset;
            String OutLine;
            int LineAddrHigh = 0;

            // blank the memory
            for (i=0; i<520000; i++) {
              gDataBfr[i] = PApplet.parseByte(unhex("FF"));
            }
            //println("mem blank");
            //println("1D00 = " + unhex("1D00") + "  1D08 = " + unhex("1D08"));
            //println("gDataBfr[1] = " + gDataBfr[1] + " ...hex = " + hex(gDataBfr[1]));
            for (i=0;i<lines.length;i++) {
              //println(i);
              LineCode = 0;
              if (lines[i].substring(0,1).equals(":")) {
                LineDataCount = unhex(lines[i].substring(1,3)); // int(byte)
                LineAddr = unhex(lines[i].substring(3,7)) & 65535; // int
                LineCode = unhex(lines[i].substring(7,9)); // int(byte)
                LineData =lines[i].substring(9,9+LineDataCount*2); // string
                //println(lines[i] + "  Count = " + LineDataCount + "  Addr = " + LineAddr + "  Code = " + LineCode + "  Data = " + LineData);

                switch(LineCode) {
                case 0: 
                  if (LineAddrHigh >= unhex("1D00") && LineAddrHigh <= unhex("1D08")) {
                    OutAddr = (LineAddrHigh * 65536) + (LineAddr & 65535); // cshort the last one
                    //println(hex((LineAddrHigh * 65536) + (LineAddr & 65535)) + " ?= " + hex((LineAddr & 65535)));
                    //println("code 0 (between) OutAddr = " + OutAddr + "  hex = " + hex(OutAddr));

                    for (int j=0; j < LineDataCount; j++) {
                      //print("code 0 OutAddr-1D000000+j = " + (OutAddr - unhex("1D000000") + j) + "  hex = " + hex(OutAddr - unhex("1D000000") + j));
                      //println("  LineData = " + LineData + "  j = " + j + "  LineData.substring(" + (j*2) + "," + (j*2+2) + ") = " + LineData.substring(j*2,j*2+2)); 
                      gDataBfr[OutAddr - unhex("1D000000") + j] =  PApplet.parseByte(unhex(LineData.substring(j*2,j*2+2)));
                    }
                    OutAddr = OutAddr + LineDataCount;
                    if (OutAddr > gPeakAddr) {
                      gPeakAddr = OutAddr;
                    }
                  }
                  /*
                  else {
                    println(hex(LineAddr) + "   " + hex(LineAddrHigh));
                  }
                  */
                  break;
                case 1: 
                  //println("code 1");
                  break;
                case 4: 
                  LineAddrHigh = unhex(lines[i].substring(9,13)) & 65535;
                  //println("code 4  LineAddrHigh = " + LineAddrHigh + "  hex = " + hex(LineAddrHigh));
                  break;
                }
              }
            }
            //println("Done making binary!");
            //println("max = " + hex(gPeakAddr));

            /*
        // write the binary to a text file to check it
             PrintWriter output;
             output = createWriter("positions.txt");
             for (int k = 0; k < (gPeakAddr - unhex("1D000000")); k++) {
             output.println(hex(k + unhex("1D000000")) + "   " + hex(gDataBfr[k]).substring(6,8));
             }
             output.flush(); // Write the remaining data
             output.close(); // Finish the file
             */

            // send program character
            myPort.write('P');
            //println("Written P");

            // wait for echo
            while (!SendBinary) {
              messageText = "Wait for P";
            }
            //println("got p");

            // wait for first G
            while (!WaitForProgrammingResponse) {
              messageText = "Wait for Write";
            }
            WaitForProgrammingResponse = false;
            messageText = "Writing";
            //println("writing");

            // send binary data in packets of 512 bytes, wait for response between packets
            int j = 0;
            int k = 0;
            while (j < (gPeakAddr - unhex("1D000000"))) {
              WaitForProgrammingResponse = false;
              for (i=0;i<16;i++) { // 32*16=512
                for (k=0;k<32;k++) {
                  dataToSend[k] = gDataBfr[j];
                  j++;
                }
                myPort.write(dataToSend);
                //myPort.write(gDataBfr[j]);
                //j++;
              }
              //println("sent j = " + j);
              //delay(150);
              //println("waiting for G, j = " + j);

              //println(num + "  " + den + "  " + num/den + "  " + progressBarLength);
              while (!WaitForProgrammingResponse) {
                messageText = "Wait for Write";
              }
              messageText = "Writing";
            }
            //println("waiting for program confirm");
            while (!WaitForProgrammingConfirm) {
              messageText = "Wait for Complete";
            }
            //println("done programming");
            messageText = "Programming Complete";
            //println("complete");

            // done programming
            SendBinary = false;
            WaitForProgrammingConfirm = false;

            // reset the NU32v2
            myPort.write('R');
            //println("sent r");

            try {
              // Clear the buffer, or available() will still be > 0
              myPort.clear();
              // Close the port
              myPort.stop();
            }
            catch (NullPointerException e) {
              println("nullpointer stopping serial");
            }
            SerialOK = false;
            ConnectedOK = false;
            //println("myport stopped");
            
            // delay seems to prevent crashes
            delay(500);

            exit();

            cursor(ARROW);
          }
          catch (NullPointerException e) {
            println("NullPointerException in Program");
            exit();
          }
        }
        /////////////////////////////// Verify //////////////////////////////////////////////////////
        if (buttonText.equals("Verify")) {
          // send verify character
          myPort.write('V');
        }
        /////////////////////////////// Run //////////////////////////////////////////////////////
        if (buttonText.equals("Run")) {
          // send run character
          myPort.write('R');

          // Clear the buffer, or available() will still be > 0
          myPort.clear();
          // Close the port
          myPort.stop();
          SerialOK = false;
          ConnectedOK = false;
          serialPortLabel = "No serial port selected";
          connectedLabel = "Not connected to NU32v2";
          messageText = "(Disconnected)";
        }
        /////////////////////////////// Exit //////////////////////////////////////////////////////
        if (buttonText.equals("Exit")) {
          exit();
        }
      }

      /////////////////////////////// Serial Button //////////////////////////////////////////////////////
      if (serialButton) {
        myPort = new Serial(this, buttonText, 115200);
        myPort.bufferUntil('\n'); 
        serialPortLabel = "'" + buttonText + "' selected";
        SerialOK = true;
      }
    }
  }
  catch (NullPointerException e) {
    println("NullPointerException in MousePressed");
    exit();
  }
}

//////////////////////////////////////////////////////////////////////////////////////
////////////  serial event   /////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

public void serialEvent(Serial thisPort) { 
  try {
    if (SerialOK) {
      // read the serial buffer as a string until a newline appears
      String myString = thisPort.readStringUntil('\n');

      // if you got any bytes other than the newline
      if (myString != null) {

        myString = trim(myString); // ditch the newline
        //println("got: " + myString);// + "  dec(1) = " + byte(myString.substring(0,1)));

        if ((myString.equals("1.1"))&&(!ConnectedOK)) {
          ConnectedOK = true; 
          connectedLabel = "NU32v2 Connected";
          messageText = "Bootloader v1.1";
        }
        else if (myString.equals("BS")) {
          messageText = "Memory Blank";
        }
        else if (myString.equals("BF")) {
          messageText = "Not Blank";
        }
        else if (myString.equals("EC")) {
          messageText = "Erase Complete";
        }
        else if (myString.equals("P")) {
          SendBinary = true;
        }
        else if (myString.equals("G")) {
          WaitForProgrammingResponse = true;
        }
        else if (myString.equals("C")&&SendBinary) {
          WaitForProgrammingConfirm = true;
        }
      }
    }
  }
  catch (NullPointerException e) {
    println("nullpointer in serial");
  }
} // end serialEvent

//////////////////////////////////////////////////////////////////////////////////////
////////////  exit   /////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////


public void stop() {
  //println("inside stop");
  try {
    if (SerialOK) {
      // Clear the buffer, or available() will still be > 0
      myPort.clear();
      // Close the port
      myPort.stop();
    }
    super.stop();
    //println("exiting...");
  }
  catch (NullPointerException e) {
    println("NullPointerException in stop");
  }
} 


//////////////////////////////////////////////////////////////////////////////////////
////////////  button class  //////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////

class Button
{
  int x, y;
  int w, h;
  int basecolor, highlightcolor, clickcolor;
  PFont thefont;
  boolean over = false;
  boolean pressed = false;   

  public void pressed() {
    if(over && mousePressed) {
      pressed = true;
    } 
    else {
      pressed = false;
    }
  }

  public boolean overRect(int x, int y, int width, int height) {
    if (mouseX >= x && mouseX <= x+width && 
      mouseY >= y && mouseY <= y+height) {
      return true;
    } 
    else {
      return false;
    }
  }
}

class TextButtons extends Button 
{
  String thetext;
  int bgcolor;

  TextButtons(int ix, int iy, int iw, int ih, PFont ithefont, String itext) 
  {
    x = ix;
    y = iy;
    w = iw;
    h = ih;
    thefont = ithefont;
    thetext = itext;
    basecolor = color(52,173,232);
    highlightcolor = color(0,102,153);
    clickcolor = color(152,200,224);//(10,240,100);
  }

  public void update() 
  {
    over();
    pressed();
    if(pressed) {
      bgcolor = clickcolor;
    } 
    else if (over) {
      bgcolor = highlightcolor;
    } 
    else {
      bgcolor = basecolor;
    }
  }

  public boolean over() 
  {
    if( overRect(x, y, w, h) ) {
      over = true;
      return true;
    } 
    else {
      over = false;
      return false;
    }
  }

  public void display() 
  {
    textFont(thefont);
    fill(bgcolor);
    rect(x,y,w,h);
    fill(255,255,255);
    text(thetext,x+5,y+h-10);
  }
}

  static public void main(String args[]) {
    PApplet.main(new String[] { "--bgcolor=#F0F0F0", "NU32v2_serial_bootloader" });
  }
}
